'''
    Apache License
                           Version 2.0, January 2004
                        http://www.apache.org/licenses/

   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

   1. Definitions.

      "License" shall mean the terms and conditions for use, reproduction,
      and distribution as defined by Sections 1 through 9 of this document.

      "Licensor" shall mean the copyright owner or entity authorized by
      the copyright owner that is granting the License.

      "Legal Entity" shall mean the union of the acting entity and all
      other entities that control, are controlled by, or are under common
      control with that entity. For the purposes of this definition,
      "control" means (i) the power, direct or indirect, to cause the
      direction or management of such entity, whether by contract or
      otherwise, or (ii) ownership of fifty percent (50%) or more of the
      outstanding shares, or (iii) beneficial ownership of such entity.

      "You" (or "Your") shall mean an individual or Legal Entity
      exercising permissions granted by this License.

      "Source" form shall mean the preferred form for making modifications,
      including but not limited to software source code, documentation
      source, and configuration files.

      "Object" form shall mean any form resulting from mechanical
      transformation or translation of a Source form, including but
      not limited to compiled object code, generated documentation,
      and conversions to other media types.

      "Work" shall mean the work of authorship, whether in Source or
      Object form, made available under the License, as indicated by a
      copyright notice that is included in or attached to the work
      (an example is provided in the Appendix below).

      "Derivative Works" shall mean any work, whether in Source or Object
      form, that is based on (or derived from) the Work and for which the
      editorial revisions, annotations, elaborations, or other modifications
      represent, as a whole, an original work of authorship. For the purposes
      of this License, Derivative Works shall not include works that remain
      separable from, or merely link (or bind by name) to the interfaces of,
      the Work and Derivative Works thereof.

      "Contribution" shall mean any work of authorship, including
      the original version of the Work and any modifications or additions
      to that Work or Derivative Works thereof, that is intentionally
      submitted to Licensor for inclusion in the Work by the copyright owner
      or by an individual or Legal Entity authorized to submit on behalf of
      the copyright owner. For the purposes of this definition, "submitted"
      means any form of electronic, verbal, or written communication sent
      to the Licensor or its representatives, including but not limited to
      communication on electronic mailing lists, source code control systems,
      and issue tracking systems that are managed by, or on behalf of, the
      Licensor for the purpose of discussing and improving the Work, but
      excluding communication that is conspicuously marked or otherwise
      designated in writing by the copyright owner as "Not a Contribution."

      "Contributor" shall mean Licensor and any individual or Legal Entity
      on behalf of whom a Contribution has been received by Licensor and
      subsequently incorporated within the Work.

   2. Grant of Copyright License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      copyright license to reproduce, prepare Derivative Works of,
      publicly display, publicly perform, sublicense, and distribute the
      Work and such Derivative Works in Source or Object form.

   3. Grant of Patent License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      (except as stated in this section) patent license to make, have made,
      use, offer to sell, sell, import, and otherwise transfer the Work,
      where such license applies only to those patent claims licensable
      by such Contributor that are necessarily infringed by their
      Contribution(s) alone or by combination of their Contribution(s)
      with the Work to which such Contribution(s) was submitted. If You
      institute patent litigation against any entity (including a
      cross-claim or counterclaim in a lawsuit) alleging that the Work
      or a Contribution incorporated within the Work constitutes direct
      or contributory patent infringement, then any patent licenses
      granted to You under this License for that Work shall terminate
      as of the date such litigation is filed.

   4. Redistribution. You may reproduce and distribute copies of the
      Work or Derivative Works thereof in any medium, with or without
      modifications, and in Source or Object form, provided that You
      meet the following conditions:

      (a) You must give any other recipients of the Work or
          Derivative Works a copy of this License; and

      (b) You must cause any modified files to carry prominent notices
          stating that You changed the files; and

      (c) You must retain, in the Source form of any Derivative Works
          that You distribute, all copyright, patent, trademark, and
          attribution notices from the Source form of the Work,
          excluding those notices that do not pertain to any part of
          the Derivative Works; and

      (d) If the Work includes a "NOTICE" text file as part of its
          distribution, then any Derivative Works that You distribute must
          include a readable copy of the attribution notices contained
          within such NOTICE file, excluding those notices that do not
          pertain to any part of the Derivative Works, in at least one
          of the following places: within a NOTICE text file distributed
          as part of the Derivative Works; within the Source form or
          documentation, if provided along with the Derivative Works; or,
          within a display generated by the Derivative Works, if and
          wherever such third-party notices normally appear. The contents
          of the NOTICE file are for informational purposes only and
          do not modify the License. You may add Your own attribution
          notices within Derivative Works that You distribute, alongside
          or as an addendum to the NOTICE text from the Work, provided
          that such additional attribution notices cannot be construed
          as modifying the License.

      You may add Your own copyright statement to Your modifications and
      may provide additional or different license terms and conditions
      for use, reproduction, or distribution of Your modifications, or
      for any such Derivative Works as a whole, provided Your use,
      reproduction, and distribution of the Work otherwise complies with
      the conditions stated in this License.

   5. Submission of Contributions. Unless You explicitly state otherwise,
      any Contribution intentionally submitted for inclusion in the Work
      by You to the Licensor shall be under the terms and conditions of
      this License, without any additional terms or conditions.
      Notwithstanding the above, nothing herein shall supersede or modify
      the terms of any separate license agreement you may have executed
      with Licensor regarding such Contributions.

   6. Trademarks. This License does not grant permission to use the trade
      names, trademarks, service marks, or product names of the Licensor,
      except as required for reasonable and customary use in describing the
      origin of the Work and reproducing the content of the NOTICE file.

   7. Disclaimer of Warranty. Unless required by applicable law or
      agreed to in writing, Licensor provides the Work (and each
      Contributor provides its Contributions) on an "AS IS" BASIS,
      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
      implied, including, without limitation, any warranties or conditions
      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
      PARTICULAR PURPOSE. You are solely responsible for determining the
      appropriateness of using or redistributing the Work and assume any
      risks associated with Your exercise of permissions under this License.

   8. Limitation of Liability. In no event and under no legal theory,
      whether in tort (including negligence), contract, or otherwise,
      unless required by applicable law (such as deliberate and grossly
      negligent acts) or agreed to in writing, shall any Contributor be
      liable to You for damages, including any direct, indirect, special,
      incidental, or consequential damages of any character arising as a
      result of this License or out of the use or inability to use the
      Work (including but not limited to damages for loss of goodwill,
      work stoppage, computer failure or malfunction, or any and all
      other commercial damages or losses), even if such Contributor
      has been advised of the possibility of such damages.

   9. Accepting Warranty or Additional Liability. While redistributing
      the Work or Derivative Works thereof, You may choose to offer,
      and charge a fee for, acceptance of support, warranty, indemnity,
      or other liability obligations and/or rights consistent with this
      License. However, in accepting such obligations, You may act only
      on Your own behalf and on Your sole responsibility, not on behalf
      of any other Contributor, and only if You agree to indemnify,
      defend, and hold each Contributor harmless for any liability
      incurred by, or claims asserted against, such Contributor by reason
      of your accepting any such warranty or additional liability.

   END OF TERMS AND CONDITIONS

   APPENDIX: How to apply the Apache License to your work.

      To apply the Apache License to your work, attach the following
      boilerplate notice, with the fields enclosed by brackets "[]"
      replaced with your own identifying information. (Don't include
      the brackets!)  The text should be enclosed in the appropriate
      comment syntax for the file format. We also recommend that a
      file or class name and description of purpose be included on the
      same "printed page" as the copyright notice for easier
      identification within third-party archives.

   Copyright [2022] [Sloimay]

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
'''


'''
BY SLOIMAY 2022 :D
Based on Fearless' implementation that he took from someone else
that he doesn't remember the name of

Versions (last is current):
1 - 15/06/2022
2 - 18/06/2022
3 - 28/06/2022
4 - 30/06/2022
5 - 10/08/2022
'''
from enum import Enum
from typing import Union

import nbtlib
from nbtlib import File
from nbtlib.tag import *

"""
Represents a minecraft schematic.
"""
class MCSchematic:

    ### --- Fields

    # The block palette used in this schematic
    # The key is the block nbt representation and the content is the
    # id. Air is the first id in the palette, for convenience.
    _blockPalette: {Union[str, int]: Union[str, int]} = {'minecraft:air': 0, 0: 'minecraft:air'}
    _blockPaletteLatestId: int = 1

    # HashMap of all the blocks, with the key being their position,
    # and the content being the id of the block in the block palette.
    # We don't want to populate _blocks with strings otherwise it's going
    # to use an ungodly amount of RAM for nothing
    _blocks: {tuple[int, int, int]: int} = {}

    # The HashMap of block entities in this schem. We store the string representation
    # for each of them instead of a fancy ID system.
    # We know we have a block entity when the inputed blockNbt has a { or }.
    # The tuple key is the position in the schem, and the value is the block state str.
    _blockEntities: {tuple[int, int, int]: str} = []

    ### ---



    ### --- Init

    def __init__(self):
        self._blockPalette = {'minecraft:air': 0, 0: 'minecraft:air'}
        self._blockPaletteLatestId = 1
        self._blocks = {}
        self._blockEntities = {}

    ### ---



    ### --- Public methods

    '''
    Saves this schematic to the inputed folder path.
    '''
    def save(self, outputFolderPath: str, schemName: str, version: 'Version'):
        # Setup
        schemMinsAndMaxes = self._getSchematicMinAndMaxXYZ()
        schemDims = self._getSchematicDimensions(schemMinsAndMaxes)
        schemOffset = self._getSchematicOffsetFrom0XYZ(schemMinsAndMaxes)


        ## BLOCK DATA
        # The block data encoded into the schematic
        encodedBlockData = bytearray()
        # Populate the encoded block data, with 0s where there
        # is no blocks placed
        for y in range(schemDims[1]):
            for z in range(schemDims[2]):
                for x in range(schemDims[0]):
                    # XYZ here is centered at 0, 0, 0 while our hashmap isn't,
                    # so we offset XYZ to match where the schem's world is.
                    positionInBlocksHashMap = (schemOffset[0] + x, schemOffset[1] + y, schemOffset[2] + z)

                    # Testing if this position is in the blocks hashmap,
                    # meaning that there is a block set at this position and not
                    # just air

                    #airId = self._blockPalette['minecraft:air']
                    #encodedBlockData.append(self._blocks.get(positionInBlocksHashMap, airId))

                    if positionInBlocksHashMap in self._blocks:
                        encodedBlockData.append(self._blocks[positionInBlocksHashMap])
                    else:
                        encodedBlockData.append(self._blockPalette['minecraft:air'])

        ## BLOCK ENTITIES
        blockEntitiesCompounds = \
            [
                self._blockEntityStringToCompound(
                    (
                        position[0] - schemOffset[0],
                        position[1] - schemOffset[1],
                        position[2] - schemOffset[2]
                    ), blockEntityStr
                )
                for position, blockEntityStr in self._blockEntities.items()
            ]


        # Get a cleaner version of the block palette, without the duplicates used
        # for back and forth O(1) access
        cleanBlockPalette = \
            {
                blockData: Int(blockPaletteId)
                for blockData, blockPaletteId in self._blockPalette.items()
                if type(blockData) == str
            }


        # Generate the schematic file from the byte array
        # From fearless's code which was taken from someone else lOl
        schematic = File({

            'Version': Int(2),
            'DataVersion': Int(version.value),
            'Metadata': Compound({
                'WEOffsetX': Int(schemOffset[0]),
                'WEOffsetY': Int(schemOffset[1]),
                'WEOffsetZ': Int(schemOffset[2]),
                'MCSchematicMetadata': Compound({
                    'Generated': String("Generated with love using Sloimay's MCSchematic Python Library, "
                                        "itself dependant on Valentin Berlier's nbtlib library.")
                })
            }),

            'Height': Short(schemDims[1]),
            'Length': Short(schemDims[2]),
            'Width': Short(schemDims[0]),

            'PaletteMax': Int(len(cleanBlockPalette)),
            'Palette': Compound(cleanBlockPalette),
            'BlockData': ByteArray(encodedBlockData),

            'BlockEntities': List(blockEntitiesCompounds),

        }, gzipped=True, root_name='Schematic')

        # DEBUG
        # print({blockNbt: Int(paletteId) for blockNbt, paletteId in self._blockPalette.items()})
        # print(schemDims)
        # print(self._blockPaletteLatestId)
        # print(self._blocks)
        # print(self._blockPalette)
        # print(self._blockEntities)
        # print(cleanBlockPalette)
        # print(len(cleanBlockPalette))
        # print(blockEntitiesCompounds)
        # print(List(blockEntitiesCompounds))

        # Save the schematic on the compuper, if no output folder is specified,
        # remove the "/", otherwise the name is gonna be "/<name>" which doesn't
        # work lOl
        schematicPath = \
            schemName + ".schem" if outputFolderPath == '' else \
            outputFolderPath + "/" + schemName + ".schem"
        schematic.save(schematicPath)

    '''
    Sets a block in the schematic at the inputed position.
    '''
    def setBlock(self, position: tuple[int, int, int], blockData: str):
        # We have a different treatment between block entities
        # and normal blocks.
        if not self._isBlockDataBlockEntity(blockData):
            self._setBlockState(position, blockData)
        else:
            self._setBlockEntity(position, blockData)

    '''
    Returns the blockState at the inputed position, None if there isn't.
    '''
    def getBlockStateAt(self, position: tuple[int, int, int]) -> Union[str, None]:
        # If there isn't any blocks at that position, return none
        if position in self._blocks:
            blockPaletteId = self._blocks[position]
            return self._blockPalette[blockPaletteId]
        else:
            return None

    ### ---



    ### --- Private methods

    '''
    Returns a new ID for the block palette when called. Always returns an ID
    that has never been used in the palette.
    '''
    def _getNewPaletteId(self) -> int:
        newId = self._blockPaletteLatestId
        self._blockPaletteLatestId += 1
        return newId

    '''
    Computes the min and max XYZs of the schematic's block positions
    '''
    def _getSchematicMinAndMaxXYZ(self) -> tuple[int, int, int, int, int, int]:
        # Setup
        minX, minY, minZ = 0, 0, 0
        maxX, maxY, maxZ = 0, 0, 0

        # Iterate through each block
        for iterationIndex, (position, blockId) in enumerate(self._blocks.items()):
            # The first iteration, we setup the min and max XYZ
            # I can't init them to be all 0s as there might be
            # blocks in the negatives or something etc etc
            if iterationIndex == 0:
                minX, minY, minZ = position
                maxX, maxY, maxZ = position

            # Finding the mins and maxes
            minX = min(minX, position[0])
            minY = min(minY, position[1])
            minZ = min(minZ, position[2])

            maxX = max(maxX, position[0])
            maxY = max(maxY, position[1])
            maxZ = max(maxZ, position[2])

        # After the loop is done we have all the mins and maxes we need
        return minX, minY, minZ, maxX, maxY, maxZ

    '''
    Gets the length in X, Y and Z of this schematic based on the mins and maxes of
    the block positions in the schem.
    '''
    def _getSchematicDimensions(self, schematicMinsAndMaxes: tuple[int, int, int, int, int, int]) -> tuple[
        int, int, int]:
        return schematicMinsAndMaxes[3] - schematicMinsAndMaxes[0] + 1, \
               schematicMinsAndMaxes[4] - schematicMinsAndMaxes[1] + 1, \
               schematicMinsAndMaxes[5] - schematicMinsAndMaxes[2] + 1

    '''
    Gets the offset of the schematic from 0, 0, 0
    Needed because we need to offset the schematic for its min XYZ to be at 0, 0, 0
    '''
    def _getSchematicOffsetFrom0XYZ(self, schematicMinsAndMaxes: tuple[int, int, int, int, int, int]) -> tuple[
        int, int, int]:
        return schematicMinsAndMaxes[0], schematicMinsAndMaxes[1], schematicMinsAndMaxes[2]

    '''
    Returns the blockData associated to the inputed ID in the palette.
    Returns None if no blockData is attached to the inputed blockPaletteId
    
    Still in the code even though it's not used in case I remove the
    duplicates in the hashmap of the block palette
    '''
    def _getNbtFromIdLinear(self, blockPaletteId: int) -> Union[str, None]:
        # Using linear search as elements in a hashmap aren't ordered
        for k, v in self._blockPalette.items():
            if v == blockPaletteId:
                return k
        return None

    '''
    Returns a Compound object from a blockEntity string, for example:
    "minecraft:barrel[]{CustomName:'{"italic":false,"text":"1"}',Items:[{Count:1b,Slot:0b,id:"minecraft:milk_bucket"}]}"
    Will get converted into a compound tag
    '''
    def _blockEntityStringToCompound(self, blockEntityPosition: tuple[int, int, int], blockEntityString: str) -> Compound:
        # The block entity nbt we are going to put in the schematic
        blockEntityNBTOut = Compound()

        # Get the NBT portion of the BEString, we know for sure that there is a '{'
        # in the string, as otherwise it would be a block iirc
        firstCurlyBracketIndex = blockEntityString.find('{')
        nbtPortion = blockEntityString[firstCurlyBracketIndex:]
        # Parse that nbt into a compound
        nbtCompound = nbtlib.parse_nbt(nbtPortion)
        # Add the parsed compound to the main compound
        blockEntityNBTOut.merge(nbtCompound)

        # Add the position to the main compound
        posCompound = Compound({'Pos': IntArray(blockEntityPosition)})
        blockEntityNBTOut.merge(posCompound)

        # Add the id to the main compound, the id may or may not have "[]" at the end of
        # it, so we checkin for that
        firstSquareBracketIndex = blockEntityString.find('[')
        if firstSquareBracketIndex == -1 or firstCurlyBracketIndex < firstSquareBracketIndex:
            # We don't have a [], because either we didn't find a "[" char, or
            # the first "[" char is after the first "{" which would mean there is
            # no "[" before "{" meaning the blockNbt's NBT starts directly after the ID,
            # without blockState tags
            idCompound = Compound({'Id': String(blockEntityString[:firstCurlyBracketIndex])})
        else:
            # We have a "[]"
            idCompound = Compound({'Id': String(blockEntityString[:firstSquareBracketIndex])})
        blockEntityNBTOut.merge(idCompound)

        # We're finished pOG
        return blockEntityNBTOut

    '''
    Returns True if the inputed blockNbt is a blockEntity.
    We check that by checking is there is at least a curly bracket in the string.
    '''
    def _isBlockDataBlockEntity(self, blockData: str) -> bool:
        return '{' in blockData

    '''
    Removes the NBT part from a blockEntityString, for example, "barrel[]{}" would
    then become "barrel[]". We assume that the blockNBT passed in has
    NBTs.
    '''
    def _getBlockStateFromBlockEntityString(self, blockEntityString: str) -> str:
        firstCurlyBracketIndex = blockEntityString.find('{')
        return blockEntityString[:firstCurlyBracketIndex]

    '''
    Returns a cleaned version of the block palette and some new mappings for it.
    For example, if we have 2 block entities that have the same blockstate but not same NBT
    such as: "barrel[]{CustomName:"a"}" and "barrel[]{CustomName:"b"}", one has the ID 0 and the
    other has an ID of 1.
    Yet in the cleaned palette, both are gonna get compiled to "barrel[]" which will have the same ID.
    So the new mappings will map old IDs that are now unused to the new one. (So in this case, every 0
    would be mapped to 1)
    
    Ended up not being implemented because I wanted to rewrite how block entities were handled  
    '''
    def _getCleanBlockPaletteAndReturnNewIdMappings(self):
        pass

    '''
    Sets the block at the inputted position. The inputted block has to be a normal block state
    '''
    def _setBlockState(self, position: tuple[int, int, int], blockData: str):
        # If the block is in the palette use its associated int, otherwise
        # add a new entry in the palette
        if blockData not in self._blockPalette:
            newPaletteId = self._getNewPaletteId()
            # Put the new block nbt in the palette
            self._blockPalette[blockData] = newPaletteId
            # Put the inverse in the palette as well when we want to
            # access in O(1) the nbt of a block from its id
            self._blockPalette[newPaletteId] = blockData

        # Set the block in this schematic to the block we just inputted
        self._blocks[position] = self._blockPalette[blockData]

    '''
    Sets the block entity at the inputed position.
    '''
    def _setBlockEntity(self, position: tuple[int, int, int], blockEntityData: str):
        # Get the block state of this blockEntity, because we still need to
        # place the block state in the world
        blockEntityState = self._getBlockStateFromBlockEntityString(blockEntityData)
        # Set the block state in the world
        self._setBlockState(position, blockEntityState)

        # Add this block entity to the block entities hashmap
        self._blockEntities[position] = blockEntityData

    ### ---



    ### --- Defaults

    '''
    Really costly repr method, as everytime you call it, you have to recompute the
    min and max XYZ of the schematic.
    '''
    def __repr__(self):
        ## Setup
        dimensions = self._getSchematicDimensions(self._getSchematicMinAndMaxXYZ())
        # The number of block states in the _blocks HashMap
        blockStateCount = len(self._blocks) // 2
        # The number of block entities
        blockEntityCount = len(self._blockEntities)

        return f"[ Dimensions={dimensions}, BlockStateCount={blockStateCount}, BlockEntityCount={blockEntityCount} ]"

    ### ---
























"""  ---- ADDITIONAL CLASSES ---- """

class QOL:
    """
    A class which holds a lot of useful quality of life methods.
    """

    ### Fields

    # A dictionary linking a signal strength to a barrel nbt string that
    # outputs this signal strength.
    _signalStrengthToSSBarrel: {int: str} = {
        0: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"0"}',Items:[]}""",
        1: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"1"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"}]}""",
        2: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"2"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"}]}""",
        3: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"3"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"}]}""",
        4: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"4"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"}]}""",
        5: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"5"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"}]}""",
        6: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"6"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"}]}""",
        7: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"7"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"}]}""",
        8: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"8"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"}]}""",
        9: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"9"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"}]}""",
        10: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"10"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"}]}""",
        11: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"11"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"},{Count:64b,Slot:18b,id:"minecraft:redstone"},{Count:64b,Slot:19b,id:"minecraft:redstone"}]}""",
        12: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"12"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"},{Count:64b,Slot:18b,id:"minecraft:redstone"},{Count:64b,Slot:19b,id:"minecraft:redstone"},{Count:64b,Slot:20b,id:"minecraft:redstone"},{Count:64b,Slot:21b,id:"minecraft:redstone"}]}""",
        13: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"13"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"},{Count:64b,Slot:18b,id:"minecraft:redstone"},{Count:64b,Slot:19b,id:"minecraft:redstone"},{Count:64b,Slot:20b,id:"minecraft:redstone"},{Count:64b,Slot:21b,id:"minecraft:redstone"},{Count:64b,Slot:22b,id:"minecraft:redstone"},{Count:64b,Slot:23b,id:"minecraft:redstone"}]}""",
        14: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"14"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"},{Count:64b,Slot:18b,id:"minecraft:redstone"},{Count:64b,Slot:19b,id:"minecraft:redstone"},{Count:64b,Slot:20b,id:"minecraft:redstone"},{Count:64b,Slot:21b,id:"minecraft:redstone"},{Count:64b,Slot:22b,id:"minecraft:redstone"},{Count:64b,Slot:23b,id:"minecraft:redstone"},{Count:64b,Slot:24b,id:"minecraft:redstone"},{Count:64b,Slot:25b,id:"minecraft:redstone"}]}""",
        15: """minecraft:barrel[open=false,facing=up]{CustomName:'{"italic":false,"text":"15"}',Items:[{Count:64b,Slot:0b,id:"minecraft:redstone"},{Count:64b,Slot:1b,id:"minecraft:redstone"},{Count:64b,Slot:2b,id:"minecraft:redstone"},{Count:64b,Slot:3b,id:"minecraft:redstone"},{Count:64b,Slot:4b,id:"minecraft:redstone"},{Count:64b,Slot:5b,id:"minecraft:redstone"},{Count:64b,Slot:6b,id:"minecraft:redstone"},{Count:64b,Slot:7b,id:"minecraft:redstone"},{Count:64b,Slot:8b,id:"minecraft:redstone"},{Count:64b,Slot:9b,id:"minecraft:redstone"},{Count:64b,Slot:10b,id:"minecraft:redstone"},{Count:64b,Slot:11b,id:"minecraft:redstone"},{Count:64b,Slot:12b,id:"minecraft:redstone"},{Count:64b,Slot:13b,id:"minecraft:redstone"},{Count:64b,Slot:14b,id:"minecraft:redstone"},{Count:64b,Slot:15b,id:"minecraft:redstone"},{Count:64b,Slot:16b,id:"minecraft:redstone"},{Count:64b,Slot:17b,id:"minecraft:redstone"},{Count:64b,Slot:18b,id:"minecraft:redstone"},{Count:64b,Slot:19b,id:"minecraft:redstone"},{Count:64b,Slot:20b,id:"minecraft:redstone"},{Count:64b,Slot:21b,id:"minecraft:redstone"},{Count:64b,Slot:22b,id:"minecraft:redstone"},{Count:64b,Slot:23b,id:"minecraft:redstone"},{Count:64b,Slot:24b,id:"minecraft:redstone"},{Count:64b,Slot:25b,id:"minecraft:redstone"},{Count:64b,Slot:26b,id:"minecraft:redstone"}]}""",
    }

    ###


    """
    Returns the blockData string of a barrel that outputs the inputted signal strength.
    If the signal strength isn't in the range [0; 15], the output is just "minecraft:barrel".
    """
    @staticmethod
    def signalStrengthToBarrelBlockDataString(signalStrength: int) -> str:
        return QOL._signalStrengthToSSBarrel.get(signalStrength, "minecraft:barrel")





class Version(Enum):
    JE_1_19_2 = 3120
    JE_1_19_2_RELEASE_CANDIDATE_2 = 3119
    JE_1_19_2_RELEASE_CANDIDATE_1 = 3118
    JE_1_19_1 = 3117
    JE_1_19_1_RELEASE_CANDIDATE_3 = 3116
    JE_1_19_1_RELEASE_CANDIDATE_2 = 3115
    JE_1_19_1_PRE_RELEASE_6 = 3114
    JE_1_19_1_PRE_RELEASE_5 = 3113
    JE_1_19_1_PRE_RELEASE_4 = 3112
    JE_1_19_1_PRE_RELEASE_3 = 3111
    JE_1_19_1_PRE_RELEASE_2 = 3110
    JE_1_19_1_RELEASE_CANDIDATE_1 = 3109
    JE_1_19_1_PRE_RELEASE_1 = 3107
    JE_22W24A = 3106
    JE_1_19 = 3105
    JE_1_19_RELEASE_CANDIDATE_2 = 3104
    JE_1_19_RELEASE_CANDIDATE_1 = 3103
    JE_1_19_PRE_RELEASE_5 = 3102
    JE_1_19_PRE_RELEASE_4 = 3101
    JE_1_19_PRE_RELEASE_3 = 3100
    JE_1_19_PRE_RELEASE_2 = 3099
    JE_1_19_PRE_RELEASE_1 = 3098
    JE_22W19A = 3096
    JE_22W18A = 3095
    JE_22W17A = 3093
    JE_22W16B = 3092
    JE_22W16A = 3091
    JE_22W15A = 3089
    JE_22W14A = 3088
    JE_22W13A = 3085
    JE_22W12A = 3082
    JE_22W11A = 3080
    JE_DEEP_DARK_EXPERIMENTAL_SNAPSHOT_1 = 3066
    JE_1_18_2 = 2975
    JE_1_18_2_RELEASE_CANDIDATE_1 = 2974
    JE_1_18_2_PRE_RELEASE_3 = 2973
    JE_1_18_2_PRE_RELEASE_2 = 2972
    JE_1_18_2_PRE_RELEASE_1 = 2971
    JE_22W07A = 2969
    JE_22W06A = 2968
    JE_22W05A = 2967
    JE_22W03A = 2966
    JE_1_18_1 = 2865
    JE_1_18_1_RELEASE_CANDIDATE_3 = 2864
    JE_1_18_1_RELEASE_CANDIDATE_2 = 2863
    JE_1_18_1_RELEASE_CANDIDATE_1 = 2862
    JE_1_18_1_PRE_RELEASE_1 = 2861
    JE_1_18 = 2860
    JE_1_18_RELEASE_CANDIDATE_4 = 2859
    JE_1_18_RELEASE_CANDIDATE_3 = 2858
    JE_1_18_RELEASE_CANDIDATE_2 = 2857
    JE_1_18_RELEASE_CANDIDATE_1 = 2856
    JE_1_18_PRE_RELEASE_8 = 2855
    JE_1_18_PRE_RELEASE_7 = 2854
    JE_1_18_PRE_RELEASE_6 = 2853
    JE_1_18_PRE_RELEASE_5 = 2851
    JE_1_18_PRE_RELEASE_4 = 2850
    JE_1_18_PRE_RELEASE_3 = 2849
    JE_1_18_PRE_RELEASE_2 = 2848
    JE_1_18_PRE_RELEASE_1 = 2847
    JE_21W44A = 2845
    JE_21W43A = 2844
    JE_21W42A = 2840
    JE_21W41A = 2839
    JE_21W40A = 2838
    JE_21W39A = 2836
    JE_21W38A = 2835
    JE_21W37A = 2834
    JE_1_18_EXPERIMENTAL_SNAPSHOT_7 = 2831
    JE_1_18_EXPERIMENTAL_SNAPSHOT_6 = 2830
    JE_1_18_EXPERIMENTAL_SNAPSHOT_5 = 2829
    JE_1_18_EXPERIMENTAL_SNAPSHOT_4 = 2828
    JE_1_18_EXPERIMENTAL_SNAPSHOT_3 = 2827
    JE_1_18_EXPERIMENTAL_SNAPSHOT_2 = 2826
    JE_1_18_EXPERIMENTAL_SNAPSHOT_1 = 2825
    JE_1_17_1 = 2730
    JE_1_17_1_RELEASE_CANDIDATE_2 = 2729
    JE_1_17_1_RELEASE_CANDIDATE_1 = 2728
    JE_1_17_1_PRE_RELEASE_3 = 2727
    JE_1_17_1_PRE_RELEASE_2 = 2726
    JE_1_17_1_PRE_RELEASE_1 = 2725
    JE_1_17 = 2724
    JE_1_17_RELEASE_CANDIDATE_2 = 2723
    JE_1_17_RELEASE_CANDIDATE_1 = 2722
    JE_1_17_PRE_RELEASE_5 = 2721
    JE_1_17_PRE_RELEASE_4 = 2720
    JE_1_17_PRE_RELEASE_3 = 2719
    JE_1_17_PRE_RELEASE_2 = 2718
    JE_1_17_PRE_RELEASE_1 = 2716
    JE_21W20A = 2715
    JE_21W19A = 2714
    JE_21W18A = 2713
    JE_21W17A = 2712
    JE_21W16A = 2711
    JE_21W15A = 2709
    JE_21W14A = 2706
    JE_21W13A = 2705
    JE_21W11A = 2703
    JE_21W10A = 2699
    JE_21W08B = 2698
    JE_21W08A = 2697
    JE_21W07A = 2695
    JE_21W06A = 2694
    JE_21W05B = 2692
    JE_21W05A = 2690
    JE_21W03A = 2689
    JE_20W51A = 2687
    JE_20W49A = 2685
    JE_20W48A = 2683
    JE_20W46A = 2682
    JE_20W45A = 2681
    JE_COMBAT_TEST_8C = 2707
    JE_COMBAT_TEST_8B = 2706
    JE_COMBAT_TEST_8 = 2705
    JE_COMBAT_TEST_7C = 2704
    JE_COMBAT_TEST_7B = 2703
    JE_COMBAT_TEST_7 = 2702
    JE_COMBAT_TEST_6 = 2701
    JE_1_16_5 = 2586
    JE_1_16_5_RELEASE_CANDIDATE_1 = 2585
    JE_1_16_4 = 2584
    JE_1_16_4_RELEASE_CANDIDATE_1 = 2583
    JE_1_16_4_PRE_RELEASE_2 = 2582
    JE_1_16_4_PRE_RELEASE_1 = 2581
    JE_1_16_3 = 2580
    JE_1_16_3_RELEASE_CANDIDATE_1 = 2579
    JE_1_16_2 = 2578
    JE_1_16_2_RELEASE_CANDIDATE_2 = 2577
    JE_1_16_2_RELEASE_CANDIDATE_1 = 2576
    JE_1_16_2_PRE_RELEASE_3 = 2575
    JE_1_16_2_PRE_RELEASE_2 = 2574
    JE_1_16_2_PRE_RELEASE_1 = 2573
    JE_20W30A = 2572
    JE_20W29A = 2571
    JE_20W28A = 2570
    JE_20W27A = 2569
    JE_1_16_1 = 2567
    JE_1_16 = 2566
    JE_1_16_RELEASE_CANDIDATE_1 = 2565
    JE_1_16_PRE_RELEASE_8 = 2564
    JE_1_16_PRE_RELEASE_7 = 2563
    JE_1_16_PRE_RELEASE_6 = 2562
    JE_1_16_PRE_RELEASE_5 = 2561
    JE_1_16_PRE_RELEASE_4 = 2560
    JE_1_16_PRE_RELEASE_3 = 2559
    JE_1_16_PRE_RELEASE_2 = 2557
    JE_1_16_PRE_RELEASE_1 = 2556
    JE_20W22A = 2555
    JE_20W21A = 2554
    JE_20W20B = 2537
    JE_20W20A = 2536
    JE_20W19A = 2534
    JE_20W18A = 2532
    JE_20W17A = 2529
    JE_20W16A = 2526
    JE_20W15A = 2525
    JE_20W14A = 2524
    JE_20W13B = 2521
    JE_20W13A = 2520
    JE_20W12A = 2515
    JE_20W11A = 2513
    JE_20W10A = 2512
    JE_20W09A = 2510
    JE_20W08A = 2507
    JE_20W07A = 2506
    JE_SNAPSHOT_20W06A = 2504
    JE_COMBAT_TEST_5 = 2321
    JE_COMBAT_TEST_4 = 2320
    JE_1_15_2 = 2230
    JE_1_15_2_PRE_RELEASE_2 = 2229
    JE_1_15_2_PRE_RELEASE_1 = 2228
    JE_1_15_1 = 2227
    JE_1_15_1_PRE_RELEASE_1 = 2226
    JE_1_15 = 2225
    JE_1_15_PRE_RELEASE_7 = 2224
    JE_1_15_PRE_RELEASE_6 = 2223
    JE_1_15_PRE_RELEASE_5 = 2222
    JE_1_15_PRE_RELEASE_4 = 2221
    JE_1_15_PRE_RELEASE_3 = 2220
    JE_1_15_PRE_RELEASE_2 = 2219
    JE_1_15_PRE_RELEASE_1 = 2218
    JE_19W46B = 2217
    JE_19W46A = 2216
    JE_19W45B = 2215
    JE_19W45A = 2214
    JE_19W44A = 2213
    JE_19W42A = 2212
    JE_19W41A = 2210
    JE_19W40A = 2208
    JE_19W39A = 2207
    JE_19W38B = 2206
    JE_19W38A = 2205
    JE_19W37A = 2204
    JE_19W36A = 2203
    JE_19W35A = 2201
    JE_19W34A = 2200
    JE_COMBAT_TEST_3 = 2069
    JE_COMBAT_TEST_2 = 2068
    JE_1_14_3___COMBAT_TEST = 2067
    JE_1_14_4 = 1976
    JE_1_14_4_PRE_RELEASE_7 = 1975
    JE_1_14_4_PRE_RELEASE_6 = 1974
    JE_1_14_4_PRE_RELEASE_5 = 1973
    JE_1_14_4_PRE_RELEASE_4 = 1972
    JE_1_14_4_PRE_RELEASE_3 = 1971
    JE_1_14_4_PRE_RELEASE_2 = 1970
    JE_1_14_4_PRE_RELEASE_1 = 1969
    JE_1_14_3 = 1968
    JE_1_14_3_PRE_RELEASE_4 = 1967
    JE_1_14_3_PRE_RELEASE_3 = 1966
    JE_1_14_3_PRE_RELEASE_2 = 1965
    JE_1_14_3_PRE_RELEASE_1 = 1964
    JE_1_14_2 = 1963
    JE_1_14_2_PRE_RELEASE_4 = 1962
    JE_1_14_2_PRE_RELEASE_3 = 1960
    JE_1_14_2_PRE_RELEASE_2 = 1959
    JE_1_14_2_PRE_RELEASE_1 = 1958
    JE_1_14_1 = 1957
    JE_1_14_1_PRE_RELEASE_2 = 1956
    JE_1_14_1_PRE_RELEASE_1 = 1955
    JE_1_14 = 1952
    JE_1_14_PRE_RELEASE_5 = 1951
    JE_1_14_PRE_RELEASE_4 = 1950
    JE_1_14_PRE_RELEASE_3 = 1949
    JE_1_14_PRE_RELEASE_2 = 1948
    JE_1_14_PRE_RELEASE_1 = 1947
    JE_19W14B = 1945
    JE_19W14A = 1944
    JE_19W13B = 1943
    JE_19W13A = 1942
    JE_19W12B = 1941
    JE_19W12A = 1940
    JE_19W11B = 1938
    JE_19W11A = 1937
    JE_19W09A = 1935
    JE_19W08B = 1934
    JE_19W08A = 1933
    JE_19W07A = 1932
    JE_19W06A = 1931
    JE_19W05A = 1930
    JE_19W04B = 1927
    JE_19W04A = 1926
    JE_19W03C = 1924
    JE_19W03B = 1923
    JE_19W03A = 1922
    JE_19W02A = 1921
    JE_18W50A = 1919
    JE_18W49A = 1916
    JE_18W48B = 1915
    JE_18W48A = 1914
    JE_18W47B = 1913
    JE_18W47A = 1912
    JE_18W46A = 1910
    JE_18W45A = 1908
    JE_18W44A = 1907
    JE_18W43C = 1903
    JE_18W43B = 1902
    JE_18W43A = 1901
    JE_1_13_2 = 1631
    JE_1_13_2_PRE2 = 1630
    JE_1_13_2_PRE1 = 1629
    JE_1_13_1 = 1628
    JE_1_13_1_PRE2 = 1627
    JE_1_13_1_PRE1 = 1626
    JE_18W33A = 1625
    JE_18W32A = 1623
    JE_18W31A = 1622
    JE_18W30B = 1621
    JE_18W30A = 1620
    JE_1_13 = 1519
    JE_1_13_PRE10 = 1518
    JE_1_13_PRE9 = 1517
    JE_1_13_PRE8 = 1516
    JE_1_13_PRE7 = 1513
    JE_1_13_PRE6 = 1512
    JE_1_13_PRE5 = 1511
    JE_1_13_PRE4 = 1504
    JE_1_13_PRE3 = 1503
    JE_1_13_PRE2 = 1502
    JE_1_13_PRE1 = 1501
    JE_18W22C = 1499
    JE_18W22B = 1498
    JE_18W22A = 1497
    JE_18W21B = 1496
    JE_18W21A = 1495
    JE_18W20C = 1493
    JE_18W20B = 1491
    JE_18W20A = 1489
    JE_18W19B = 1485
    JE_18W19A = 1484
    JE_18W16A = 1483
    JE_18W15A = 1482
    JE_18W14B = 1481
    JE_18W14A = 1479
    JE_18W11A = 1478
    JE_18W10D = 1477
    JE_18W10C = 1476
    JE_18W10B = 1474
    JE_18W10A = 1473
    JE_18W09A = 1472
    JE_18W08B = 1471
    JE_18W08A = 1470
    JE_18W07C = 1469
    JE_18W07B = 1468
    JE_18W07A = 1467
    JE_18W06A = 1466
    JE_18W05A = 1464
    JE_18W03B = 1463
    JE_18W03A = 1462
    JE_18W02A = 1461
    JE_18W01A = 1459
    JE_17W50A = 1457
    JE_17W49B = 1455
    JE_17W49A = 1454
    JE_17W48A = 1453
    JE_17W47B = 1452
    JE_17W47A = 1451
    JE_17W46A = 1449
    JE_17W45B = 1448
    JE_17W45A = 1447
    JE_17W43B = 1445
    JE_17W43A = 1444
    JE_1_12_2 = 1343
    JE_1_12_2_PRE2 = 1342
    JE_1_12_2_PRE1 = 1341
    JE_1_12_1 = 1241
    JE_1_12_1_PRE1 = 1240
    JE_17W31A = 1239
    JE_1_12 = 1139
    JE_1_12_PRE7 = 1138
    JE_1_12_PRE6 = 1137
    JE_1_12_PRE5 = 1136
    JE_1_12_PRE4 = 1135
    JE_1_12_PRE3 = 1134
    JE_1_12_PRE2 = 1133
    JE_1_12_PRE1 = 1132
    JE_17W18B = 1131
    JE_17W18A = 1130
    JE_17W17B = 1129
    JE_17W17A = 1128
    JE_17W16B = 1127
    JE_17W16A = 1126
    JE_17W15A = 1125
    JE_17W14A = 1124
    JE_17W13B = 1123
    JE_17W13A = 1122
    JE_17W06A = 1022
    JE_1_11_2 = 922
    JE_1_11_1 = 921
    JE_16W50A = 920
    JE_1_11 = 819
    JE_1_11_PRE1 = 818
    JE_16W44A = 817
    JE_16W43A = 816
    JE_16W42A = 815
    JE_16W41A = 814
    JE_16W40A = 813
    JE_16W39C = 812
    JE_16W39B = 811
    JE_16W39A = 809
    JE_16W38A = 807
    JE_16W36A = 805
    JE_16W35A = 803
    JE_16W33A = 802
    JE_16W32B = 801
    JE_16W32A = 800
    JE_1_10_2 = 512
    JE_1_10_1 = 511
    JE_1_10 = 510
    JE_1_10_PRE2 = 507
    JE_1_10_PRE1 = 506
    JE_16W21B = 504
    JE_16W21A = 503
    JE_16W20A = 501
    JE_1_9_4 = 184
    JE_1_9_3 = 183
    JE_1_9_3_PRE3 = 182
    JE_1_9_3_PRE2 = 181
    JE_1_9_3_PRE1 = 180
    JE_16W15B = 179
    JE_16W15A = 178
    JE_16W14A = 177
    JE_1_9_2 = 176
    JE_1_9_1 = 175
    JE_1_9_1_PRE3 = 172
    JE_1_9_1_PRE2 = 171
    JE_1_9_1_PRE1 = 170
    JE_1_9 = 169
    JE_1_9_PRE4 = 168
    JE_1_9_PRE3 = 167
    JE_1_9_PRE2 = 165
    JE_1_9_PRE1 = 164
    JE_16W07B = 163
    JE_16W07A = 162
    JE_16W06A = 161
    JE_16W05B = 160
    JE_16W05A = 159
    JE_16W04A = 158
    JE_16W03A = 157
    JE_16W02A = 156
    JE_15W51B = 155
    JE_15W51A = 154
    JE_15W50A = 153
    JE_15W49B = 152
    JE_15W49A = 151
    JE_15W47C = 150
    JE_15W47B = 149
    JE_15W47A = 148
    JE_15W46A = 146
    JE_15W45A = 145
    JE_15W44B = 143
    JE_15W44A = 142
    JE_15W43C = 141
    JE_15W43B = 140
    JE_15W43A = 139
    JE_15W42A = 138
    JE_15W41B = 137
    JE_15W41A = 136
    JE_15W40B = 134
    JE_15W40A = 133
    JE_15W39C = 132
    JE_15W39B = 131
    JE_15W39A = 130
    JE_15W38B = 129
    JE_15W38A = 128
    JE_15W37A = 127
    JE_15W36D = 126
    JE_15W36C = 125
    JE_15W36B = 124
    JE_15W36A = 123
    JE_15W35E = 122
    JE_15W35D = 121
    JE_15W35C = 120
    JE_15W35B = 119
    JE_15W35A = 118
    JE_15W34D = 117
    JE_15W34C = 116
    JE_15W34B = 115
    JE_15W34A = 114
    JE_15W33C = 112
    JE_15W33B = 111
    JE_15W33A = 55
    JE_15W32C = 104
    JE_15W32B = 103
    JE_15W32A = 100
